// This is sample code for the JRB smart pointer package

// ***********************************
// Copyright 1999 John R. Bandela
// This work may be freely copied, distributed, and extended, provided
// the copyright notice above and this statement are maintained at the top of the code


#include <iostream>
#include "jrb_ptr.h"

struct A{		
	void* operator new(size_t sz){
		std::cout << "A::new\n";
		void * p = new unsigned char[sz];
		std::cout<< p << "\n\n";
		return p;
	}
	void operator delete(void* p){
		std::cout << "A::delete\n" << p << "\n\n";
		delete p;
	}
};
struct B:public A{
	void* operator new(size_t sz){
		std::cout << "B::new\n";
		void * p = new unsigned char[sz];
		std::cout<< p << "\n\n";
		return p;
	}
	void operator delete(void* p){
		std::cout << "B::delete\n" << p << "\n\n";
		delete p;
	}
	int q;
};
struct C{};
struct D:public C{int k;};

int main(){
	using namespace jrb_sptr;
	using namespace std;
	// Declare a back pointer
	// A back pointer prevents dangling pointers to invalid objects
	// but does not keep the object from being destroyed
	BackPtr<A> back;
	{
		// Create a new ptr
		//(You have to use Make_CP to assign a raw pointer to a counted one)
		// This makes sure the correct type info gets used by the counted ptr.
		CountedPtr<B> bptr = Make_CP(new B);
		
		// Assign it to the back pointer, of a less derived class
		back = bptr;
		
		// Assign to a less derived class
		CountedPtr<A> aptr = bptr;
		// Test the number of references
		cout << "references on aptr = " << aptr.GetRefCount() << "\n";
		// Test for equality
		cout << (aptr==bptr) << "\n";
		
		// The following line will not compile
		// CountedPtr<B> b2 = aptr;
		
		// However, we can use static cast like syntax to downcast
		CountedPtr<B> b3 = CP_static_cast<B>(aptr);
		
		// Check the references
		cout << "references on aptr = " << aptr.GetRefCount() << "\n";
		cout << "references on bptr = " << bptr.GetRefCount() << "\n";
		cout << "references on b3 = " << b3.GetRefCount() << "\n";
		
		// Release bptr
		cout << "Releasing bptr\n";
		bptr = CP_NULL;
		
		// Check the references
		cout << "references on aptr = " << aptr.GetRefCount() << "\n";
		cout << "references on bptr = " << bptr.GetRefCount() << "\n";
		cout << "references on b3 = " << b3.GetRefCount() << "\n";
		
		// Now release b3
		cout << "Releasing b3\n";
		b3 = CP_NULL;
		
		// Check the references
		cout << "references on aptr = " << aptr.GetRefCount() << "\n";
		cout << "references on bptr = " << bptr.GetRefCount() << "\n";
		cout << "references on b3 = " << b3.GetRefCount() << "\n";
		
		// Output the type info for the name
		cout << aptr.GetTypeInfo().name() << "\n";
		
		// Check the validity of the back pointer (should be 1)
		cout << "Is back valid? " << back.IsValid() << "\n";
		
		// Get an CountedPtr<A> from the back pointer
		CountedPtr<A> a2 = back.GetCountedPtr();
		
		// Check the references
		cout << "references on aptr = " << aptr.GetRefCount() << "\n";
		cout << "references on bptr = " << bptr.GetRefCount() << "\n";
		cout << "references on b3 = " << b3.GetRefCount() << "\n";
		cout << "references on a2 = " << a2.GetRefCount() << "\n";
		
		// Allow aptr and a2 to go out of scope
		// The refcount should go to zero and aptr will delete the pointer
		// Notice that although aptr is type A, the correct delete (type B) will be called
		// In addition, notice that back will not prevent aptr and a2 from deleting the object
	}
	// Now check validity of back
	cout << "Is back valid? " << back.IsValid() << "\n";
	
	// The following code will produce an exception
	// CountedPtr<A> a4 = back.GetCountedPtr();
	
	// Create a new D and assign it to a C ptr
	CountedPtr<C> c1 = Make_CP(new D);
	
	// Use static cast to assign the c1 to a D ptr
	CountedPtr<D> d1 = CP_static_cast<D>(c1);
	
	// Change the value of d1->k
	d1->k = 6;
	
	// Clone c1 now
	CountedPtr<C> c2 = c1.Clone();
	
	// c1 and c2 should not be equal
	cout << "c1==c2? " << (c1==c2) << "\n";
	
	// Assign c2 to d2 using a static type cast
	CountedPtr<D> d2 = CP_static_cast<D>(c2);
	
	// d2 and c2 should  be equal
	cout << "d2==c2? " << (d2==c2) << "\n";
	
	// d2 and d1 should not be equal
	cout << "d2==d1? " << (d2==d1) << "\n";
	
	// However, d2->k and d1-> should both equal 6;
	cout << "d1->k = " << d1->k << ";  d1->k = " << d2->k << "\n";
	
	
	return 0;
}
